import logging
import os
import sys
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path

from . import assets, events, replication, time_series
from ._version import __version__


def configure_logger(log_level: str = "INFO", log_path: Path = None) -> None:
    """Configure the logging to stdout and optionally local file and GCP stackdriver."""
    log_handlers = [logging.StreamHandler(sys.stdout)]

    if log_path is not None:
        log_path.mkdir(parents=True, exist_ok=True)
        log_file = log_path.joinpath("cognite-replicator.log")
        log_handlers.append(TimedRotatingFileHandler(log_file, when="midnight", backupCount=7))

    logging.basicConfig(
        level=logging.INFO if log_level.upper() == "INFO" else log_level,
        format="%(asctime)s %(name)s %(levelname)s - %(message)s",
        handlers=log_handlers,
    )

    _configure_stackdriver_logging()


def configure_databricks_logger(
    log_level=logging.INFO, logger: logging.Logger = None, file_path: str = None
) -> logging.Logger:
    """Configure logging for databricks.

    Args:
        log_level: the logging level
        logger: the logger to use, default is root logger
        file_path: the path to a file for storing logs to persistent disk if provided
    """
    if logger is None:
        logger = logging.getLogger()
    logger.setLevel(log_level)
    log_handlers = [logging.StreamHandler(sys.stdout)]
    if file_path is not None:
        file_handler = TimedRotatingFileHandler(file_path, when="midnight", backupCount=7)
        file_handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
        log_handlers.append(file_handler)
    logger.handlers = log_handlers
    logging.getLogger("py4j").setLevel(logging.ERROR)  # To remove the unnecessary databricks logging output
    return logger


def _configure_stackdriver_logging() -> None:
    """Send logs to GCP stackdriver. Must be configured with GOOGLE_APPLICATION_CREDENTIALS."""
    if os.environ.get("GOOGLE_APPLICATION_CREDENTIALS"):
        try:
            import google.cloud.logging
        except ImportError:
            logging.warning("GOOGLE_APPLICATION_CREDENTIALS set but google-cloud-logging not available")
        else:
            google.cloud.logging.Client().setup_logging(name="cognite-replicator")



import logging
import re
import time
from typing import Dict, List, Optional

from cognite.client import CogniteClient
from cognite.client.data_classes import TimeSeries, TimeSeriesList
from cognite.client.exceptions import CogniteNotFoundError

from . import replication


def create_time_series(
    src_ts: TimeSeries, src_dst_ids_assets: Dict[int, int], project_src: str, runtime: int
) -> TimeSeries:
    """
    Make a new copy of the time series to be replicated based on a source time series.

    Args:
        src_ts: The time series from the source to be replicated to the destination.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.

    Returns:
        The replicated time series to be created in the destination.
    """
    logging.debug(f"Creating a new time series based on source time series id {src_ts.id}")

    return TimeSeries(
        external_id=src_ts.external_id,
        name=src_ts.name,
        is_string=src_ts.is_string,
        metadata=replication.new_metadata(src_ts, project_src, runtime),
        unit=src_ts.unit,
        asset_id=replication.get_asset_ids([src_ts.asset_id], src_dst_ids_assets)[0] if src_ts.asset_id else None,
        is_step=src_ts.is_step,
        description=src_ts.description,
        security_categories=src_ts.security_categories,
        legacy_name=src_ts.external_id,
    )


def update_time_series(
    src_ts: TimeSeries, dst_ts: TimeSeries, src_dst_ids_assets: Dict[int, int], project_src: str, runtime: int
) -> TimeSeries:
    """
    Makes an updated version of the destination time series based on the corresponding source time series.

    Args:
        src_ts: The time series from the source to be replicated.
        dst_ts: The time series from the destination that needs to be updated to reflect changes made to its
                source time series.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.

    Returns:
        The updated time series object for the replication destination.
    """
    logging.debug(f"Updating existing time series {dst_ts.id} based on source time series id {src_ts.id}")

    dst_ts.external_id = src_ts.external_id
    dst_ts.name = src_ts.name
    dst_ts.is_string = src_ts.is_string
    dst_ts.metadata = replication.new_metadata(src_ts, project_src, runtime)
    dst_ts.unit = src_ts.unit
    dst_ts.asset_id = replication.get_asset_ids([src_ts.asset_id], src_dst_ids_assets)[0] if src_ts.asset_id else None
    dst_ts.is_step = src_ts.is_step
    dst_ts.description = src_ts.description
    dst_ts.security_categories = src_ts.security_categories
    return dst_ts


def _has_security_category(ts: TimeSeries) -> bool:
    return ts.security_categories is not None and len(ts.security_categories) > 0


def _is_service_account_metric(ts: TimeSeries) -> bool:
    return ts.name is not None and "service_account_metrics" in ts.name


def _is_copyable(ts: TimeSeries) -> bool:
    return not _is_service_account_metric(ts) and not _has_security_category(ts)


def copy_ts(
    src_ts: List[TimeSeries],
    src_id_dst_ts: Dict[int, TimeSeries],
    src_dst_ids_assets: Dict[int, int],
    project_src: str,
    runtime: int,
    client: CogniteClient,
    dst_ts: List[TimeSeries],
):
    """
    Creates/updates time series objects and then attempts to create and update these time series in the destination.

    Args:
        src_ts: A list of the time series that are in the source.
        src_id_dst_ts: A dictionary of a time series source id to it's matching destination object.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.
        client: The client corresponding to the destination project.
        dst_ts: List of timeseries in the destination - Will be used for comparison if current timeseries where not copied by the replicator
    """
    logging.info(f"Starting to replicate {len(src_ts)} time series.")
    for ts in src_ts:  # for unlinkable time_series, remove asset id to avoid insertion error
        if ts.asset_id is not None and ts.asset_id not in src_dst_ids_assets:
            ts.asset_id = None

    create_ts, update_ts, unchanged_ts = replication.make_objects_batch(
        src_ts,
        src_id_dst_ts,
        src_dst_ids_assets,
        create_time_series,
        update_time_series,
        project_src,
        runtime,
        dst_ts=dst_ts,
    )

    logging.info(f"Creating {len(create_ts)} new time series and updating {len(update_ts)} existing time series.")

    if create_ts:
        logging.info(f"Creating {len(create_ts)} time series.")
        created_ts = replication.retry(client.time_series.create, create_ts)
        logging.info(f"Successfully created {len(created_ts)} time series.")

    if update_ts:
        logging.info(f"Updating {len(update_ts)} time series.")
        updated_ts = replication.retry(client.time_series.update, update_ts)
        logging.info(f"Successfully updated {len(updated_ts)} time series.")

    if unchanged_ts:
        logging.info(f"{len(unchanged_ts)} time series will not be changed.")


def replicate(
    client_src: CogniteClient,
    client_dst: CogniteClient,
    batch_size: int = 10000,
    num_threads: int = 1,
    delete_replicated_if_not_in_src: bool = False,
    delete_not_replicated_in_dst: bool = False,
    skip_unlinkable: bool = False,
    skip_nonasset: bool = False,
    target_external_ids: Optional[List[str]] = None,
    exclude_pattern: str = None,
):
    """
    Replicates all the time series from the source project into the destination project.

    Args:
        client_src: The client corresponding to the source project.
        client_dst: The client corresponding to the destination project.
        batch_size: The biggest batch size to post chunks in.
        num_threads: The number of threads to be used.
        delete_replicated_if_not_in_src: If True, will delete replicated assets that are in the destination,
        but no longer in the source project (Default=False).
        delete_not_replicated_in_dst: If True, will delete assets from the destination if they were not replicated
        from the source (Default=False).
        skip_unlinkable: If no assets exist in the destination for a time series, do not replicate it
        skip_nonasset: If a time series has no associated assets, do not replicate it
        target_external_ids: List of specific time series external ids to replicate,
        exclude_pattern: Regex pattern; time series whose names match will not be replicated
    """
    project_src = client_src.config.project
    project_dst = client_dst.config.project

    if target_external_ids:
        ts_src = client_src.time_series.retrieve_multiple(external_ids=target_external_ids, ignore_unknown_ids=True)
        try:
            ts_dst = client_dst.time_series.retrieve_multiple(external_ids=target_external_ids, ignore_unknown_ids=True)
        except CogniteNotFoundError:
            ts_dst = TimeSeriesList([])
    else:
        ts_src = client_src.time_series.list(limit=None)
        ts_dst = client_dst.time_series.list(limit=None)
        logging.info(f"There are {len(ts_src)} existing time series in source ({project_src}).")
        logging.info(f"There are {len(ts_dst)} existing time series in destination ({project_dst}).")

    src_id_dst_ts = replication.make_id_object_map(ts_dst)

    assets_dst = client_dst.assets.list(limit=None)
    src_dst_ids_assets = replication.existing_mapping(*assets_dst)
    logging.info(
        f"If a time series asset id is one of the {len(src_dst_ids_assets)} assets "
        f"that have been replicated then it will be linked."
    )

    compiled_re = None
    if exclude_pattern:
        compiled_re = re.compile(exclude_pattern)

    def filter_fn(ts):
        if exclude_pattern:
            return _is_copyable(ts) and compiled_re.search(ts.external_id) is None
        return _is_copyable(ts)

    ts_src_filtered = replication.filter_objects(ts_src, src_dst_ids_assets, skip_unlinkable, skip_nonasset, filter_fn)
    logging.info(
        f"Filtered out {len(ts_src) - len(ts_src_filtered)} time series. {len(ts_src_filtered)} time series remain."
    )

    replicated_runtime = int(time.time()) * 1000
    logging.info(f"These copied/updated time series will have a replicated run time of: {replicated_runtime}.")

    logging.info(
        f"Starting to copy and update {len(ts_src_filtered)} time series from "
        f"source ({project_src}) to destination ({project_dst})."
    )

    if len(ts_src_filtered) > batch_size:
        replication.thread(
            num_threads=num_threads,
            copy=copy_ts,
            src_objects=ts_src_filtered,
            src_id_dst_obj=src_id_dst_ts,
            src_dst_ids_assets=src_dst_ids_assets,
            project_src=project_src,
            replicated_runtime=replicated_runtime,
            client=client_dst,
            dst_ts=ts_dst,
        )
    else:
        copy_ts(
            src_ts=ts_src_filtered,
            src_id_dst_ts=src_id_dst_ts,
            src_dst_ids_assets=src_dst_ids_assets,
            project_src=project_src,
            runtime=replicated_runtime,
            client=client_dst,
            dst_ts=ts_dst,
        )

    logging.info(
        f"Finished copying and updating {len(ts_src_filtered)} time series from "
        f"source ({project_src}) to destination ({project_dst})."
    )

    if delete_replicated_if_not_in_src:
        ids_to_delete = replication.find_objects_to_delete_if_not_in_src(ts_src, ts_dst)
        client_dst.time_series.delete(id=ids_to_delete)
        logging.info(
            f"Deleted {len(ids_to_delete)} time series destination ({project_dst})"
            f" because they were no longer in source ({project_src})   "
        )
    if delete_not_replicated_in_dst:
        ids_to_delete = replication.find_objects_to_delete_not_replicated_in_dst(ts_dst)
        client_dst.time_series.delete(id=ids_to_delete)
        logging.info(
            f"Deleted {len(ids_to_delete)} time series in destination ({project_dst}) because"
            f"they were not replicated from source ({project_src})   "
        )



import logging
import time
from typing import Dict, List, Optional, Union

from cognite.client import CogniteClient
from cognite.client.data_classes.assets import Asset, AssetList

from . import replication


def build_asset_create(
    src_asset: Asset, src_id_dst_map: Dict[int, int], project_src: str, runtime: int, depth: int
) -> Asset:
    """
    Makes a new copy of the asset to be replicated based on the source asset.

    Args:
        src_asset: The asset from the source to be replicated to destination.
        src_id_dst_map: A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.
        depth: The depth of the asset within the asset hierarchy.

    Returns:
        The replicated asset to be created in the destination.

    """
    logging.debug(f"Creating a new asset based on source event id {src_asset.id}")

    return Asset(
        external_id=src_asset.external_id,
        name=src_asset.name,
        description=src_asset.description,
        metadata=replication.new_metadata(src_asset, project_src, runtime),
        source=src_asset.source,
        parent_id=src_id_dst_map[src_asset.parent_id] if depth > 0 else None,
    )


def build_asset_update(
    src_asset: Asset, dst_asset: Asset, src_id_dst_map: Dict[int, int], project_src: str, runtime: int, depth: int
) -> Asset:
    """
    Makes an updated version of the destination asset based on the corresponding source asset.

    Args:
        src_asset: The asset from the source to be replicated to destination.
        dst_asset: The asset from the destination that needs to be updated to reflect changes made to its source asset.
        src_id_dst_map: \\*\\*A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.
        depth: \\*\\*The depth of the asset within the asset hierarchy.
    \\*\\* only needed when hierarchy becomes mutable

    Returns:
        The updated asset object for the replication destination.

    """
    logging.debug(f"Updating existing event {dst_asset.id} based on source event id {src_asset.id}")

    dst_asset.external_id = src_asset.external_id
    dst_asset.name = src_asset.name
    dst_asset.description = src_asset.description
    dst_asset.metadata = replication.new_metadata(src_asset, project_src, runtime)
    dst_asset.source = src_asset.source
    dst_asset.parent_id = src_id_dst_map[src_asset.parent_id] if depth > 0 else None  # when asset hierarchy is mutable
    return dst_asset


def find_children(assets: List[Asset], parents: Union[List[None], List[Asset]]) -> List[Asset]:
    """
    Creates a list of all the assets that are children of the parent assets.

    Args:
        assets: A list of all the assets to search for children from.
        parents: A list of all the assets to find children for.

    Returns:
        A list of all the assets that are children to the parents.

    """
    parent_ids = {parent.id for parent in parents} if parents != [None] else parents
    return [asset for asset in assets if asset.parent_id in parent_ids]


def create_hierarchy(
    src_assets: List[Asset],
    dst_assets: List[Asset],
    project_src: str,
    runtime: int,
    client: CogniteClient,
    subtree_ids: Optional[List[int]] = None,
    subtree_external_ids: Optional[List[str]] = None,
    subtree_max_depth: Optional[int] = None,
):
    """
    Creates/updates the asset hierarchy in batches by depth, starting with the root assets and then moving on to the
    children of those roots, etc.

    Args:
        src_assets: A list of the assets that are in the source.
        dst_assets: A list of the assets that are in the destination.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.
        client: The client corresponding to the destination project.
        subtree_ids: The id of the subtree root to replicate,
        subtree_external_ids: The external id of the subtree root to replicate,
        subtree_max_depth: The maximum tree depth to replicate,
    """
    depth = 0
    parents = [None]  # root nodes parent id is None
    if subtree_ids is not None or subtree_external_ids is not None:
        unlink_subtree_parents(src_assets, subtree_ids, subtree_external_ids)
    children = find_children(src_assets, parents)

    src_dst_ids: Dict[int, int] = {}
    src_id_dst_asset = replication.make_id_object_map(dst_assets)

    while children:
        logging.info(f"Starting depth {depth}, with {len(children)} assets.")
        create_assets, update_assets, unchanged_assets = replication.make_objects_batch(
            children,
            src_id_dst_asset,
            src_dst_ids,
            build_asset_create,
            build_asset_update,
            project_src,
            runtime,
            depth=depth,
        )

        logging.info(f"Attempting to create {len(create_assets)} assets.")
        created_assets = replication.retry(client.assets.create, create_assets)
        logging.info(f"Attempting to update {len(update_assets)} assets.")
        updated_assets = replication.retry(client.assets.update, update_assets)

        src_dst_ids = replication.existing_mapping(*created_assets, *updated_assets, *unchanged_assets, ids=src_dst_ids)
        logging.debug(f"Dictionary of current asset mappings: {src_dst_ids}")

        num_assets = len(created_assets) + len(updated_assets)
        logging.info(
            f"Finished depth {depth}, updated {len(updated_assets)} and "
            f"posted {len(created_assets)} assets (total of {num_assets} assets)."
        )

        depth += 1
        if subtree_max_depth is not None and depth > subtree_max_depth:
            logging.info("Reached max depth")
            break
        children = find_children(src_assets, children)

    return src_dst_ids


def unlink_subtree_parents(
    src_assets: List[Asset], subtree_ids: Optional[List[int]] = None, subtree_external_ids: Optional[List[str]] = None
):
    """Sets the parentId of subtree roots to be None

    Args:
        src_assets: A list of the assets that are in the source.
        subtree_ids: The id(s) of the subtree root(s).
        subtree_external_ids: The external id(s) of the subtree root(s).
    """
    logging.info("Searching for subtree parent...")
    for asset in src_assets:
        if (subtree_ids is not None and asset.id in subtree_ids) or (
            subtree_external_ids is not None and asset.external_id in subtree_external_ids
        ):
            logging.info(f"Found the subtree root: {asset.id} with parent id: {asset.parent_id}")
            if asset.metadata is None:
                asset.metadata = {}

            if asset.parent_id:
                asset.metadata["_replicatedOriginalParentId"] = asset.parent_id
            if asset.parent_external_id:
                asset.metadata["_replicatedOriginalParentExternalId"] = asset.parent_external_id
            asset.parent_id = None
            asset.parent_external_id = None


def replicate(
    client_src: CogniteClient,
    client_dst: CogniteClient,
    delete_replicated_if_not_in_src: bool = False,
    delete_not_replicated_in_dst: bool = False,
    subtree_ids: Optional[Union[int, List[int]]] = None,
    subtree_external_ids: Optional[Union[str, List[str]]] = None,
    subtree_max_depth: Optional[int] = None,
):
    """
    Replicates all the assets from the source project into the destination project.

    Args:
        client_src: The client corresponding to the source project.
        client_dst: The client corresponding to the destination project.
        delete_replicated_if_not_in_src: If True, will delete replicated assets that are in the destination,
        but no longer in the source project (Default=False).
        delete_not_replicated_in_dst: If True, will delete assets from the destination if they were not replicated
        from the source (Default=False).
        subtree_ids: The id or list of ids of subtree root to replicate,
        subtree_external_ids: The external id or list of external ids of the subtree root to replicate,
        subtree_max_depth: The maximum tree depth to replicate,
    """
    project_src = client_src.config.project
    project_dst = client_dst.config.project

    if not subtree_ids and not subtree_external_ids:
        assets_src = client_src.assets.list(limit=None)
    else:
        logging.info(f"Loading subtree(s) with id(s) {subtree_ids} or external id(s) {subtree_external_ids}")
        assets_src = AssetList(resources=[])
        if subtree_ids:
            if not isinstance(subtree_ids, list):
                subtree_ids = [subtree_ids]
            for subtree_id in subtree_ids:
                assets_src += client_src.assets.retrieve_subtree(id=subtree_id, depth=subtree_max_depth)
        if subtree_external_ids:
            if not isinstance(subtree_external_ids, list):
                subtree_external_ids = [subtree_external_ids]
            for subtree_id in subtree_external_ids:
                assets_src += client_src.assets.retrieve_subtree(external_id=subtree_id, depth=subtree_max_depth)
    assets_dst = client_dst.assets.list(limit=None)

    logging.info(f"There are {len(assets_src)} existing assets in source ({project_src}).")
    logging.info(f"There are {len(assets_dst)} existing assets in destination ({project_dst}).")

    replicated_runtime = int(time.time()) * 1000
    logging.info(f"These copied/updated assets will have a replicated run time of: {replicated_runtime}.")

    logging.info(
        f"Starting to copy and update {len(assets_src)} assets from "
        f"source ({project_src}) to destination ({project_dst})."
    )
    src_dst_ids_assets = create_hierarchy(
        assets_src,
        assets_dst,
        project_src,
        replicated_runtime,
        client_dst,
        subtree_ids,
        subtree_external_ids,
        subtree_max_depth,
    )

    logging.info(
        f"Finished copying and updating {len(src_dst_ids_assets)} assets from "
        f"source ({project_src}) to destination ({project_dst})."
    )

    if delete_replicated_if_not_in_src:
        ids_to_delete = replication.find_objects_to_delete_if_not_in_src(assets_src, assets_dst)
        client_dst.assets.delete(id=ids_to_delete)
        logging.info(
            f"Deleted {len(ids_to_delete)} assets in destination ({project_dst})"
            f" because they were no longer in source ({project_src})   "
        )
    if delete_not_replicated_in_dst:
        ids_to_delete = replication.find_objects_to_delete_not_replicated_in_dst(assets_dst)
        client_dst.assets.delete(id=ids_to_delete)
        logging.info(
            f"Deleted {len(ids_to_delete)} assets in destination ({project_dst}) because"
            f"they were not replicated from source ({project_src})   "
        )



import logging
import mimetypes
import time
from typing import Dict, List

from cognite.client import CogniteClient
from cognite.client.data_classes import FileMetadata
from cognite.client.exceptions import CogniteAPIError

from . import replication


def create_file(
    src_file: FileMetadata, src_dst_ids_assets: Dict[int, int], project_src: str, runtime: int
) -> FileMetadata:
    """
    Makes a new copy of the file to be replicated based on a source file.

    Args:
        src_file: The file from the source to be replicated to destination.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.

    Returns:
        The replicated file to be created in the destination.
    """
    logging.debug(f"Creating a new file based on source file id {src_file.id}")
    mime_type = mimetypes.types_map.get(f".{src_file.mime_type}", src_file.mime_type)

    return FileMetadata(
        external_id=src_file.external_id,
        name=src_file.name,
        source=src_file.source,
        mime_type=mime_type,
        metadata=replication.new_metadata(src_file, project_src, runtime),
        asset_ids=replication.get_asset_ids(src_file.asset_ids, src_dst_ids_assets),
    )


def update_file(
    src_file: FileMetadata, dst_file: FileMetadata, src_dst_ids_assets: Dict[int, int], project_src: str, runtime: int
) -> FileMetadata:
    """
    Makes an updated version of the destination file based on the corresponding source file.

    Args:
        src_file: The file from the source to be replicated.
        dst_file: The file from the destination that needs to be updated to reflect changes made to its source file.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.

    Returns:
        The updated file object for the replication destination.
    """
    logging.debug(f"Updating existing file {dst_file.id} based on source file id {src_file.id}")

    dst_file.external_id = src_file.external_id
    dst_file.source = src_file.source
    dst_file.metadata = replication.new_metadata(src_file, project_src, runtime)
    dst_file.asset_ids = replication.get_asset_ids(src_file.asset_ids, src_dst_ids_assets)
    dst_file.source_created_time = src_file.source_created_time
    dst_file.source_modified_time = src_file.source_modified_time
    return dst_file


def copy_files(
    src_files: List[FileMetadata],
    src_id_dst_file: Dict[int, FileMetadata],
    src_dst_ids_assets: Dict[int, int],
    project_src: str,
    runtime: int,
    client: CogniteClient,
):
    """
    Creates/updates file objects and then attempts to create and update these objects in the destination.

    Args:
        src_files: A list of the files that are in the source.
        src_id_dst_file:  A dictionary of a files source id to it's matching destination object.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.
        client: The client corresponding to the destination project.

    """
    logging.debug(f"Starting to replicate {len(src_files)} files.")

    create_files, update_files, unchanged_files = replication.make_objects_batch(
        src_files, src_id_dst_file, src_dst_ids_assets, create_file, update_file, project_src, runtime
    )

    logging.info(f"Creating {len(create_files)} new files and updating {len(update_files)} existing files.")

    create_urls = []
    if create_files:
        logging.debug(f"Attempting to create {len(create_files)} files.")
        for file in create_files:
            response = None
            try:
                response = replication.retry(client.files.create, file)
            except CogniteAPIError as exc:
                logging.error(f"Failed to create file {file.name}. {exc}")
                if "Invalid MIME type" in exc.message:
                    file.mime_type = None
                    response = replication.retry(client.files.create, file)

            if response:
                create_urls.append(response)
        logging.debug(f"Successfully created {len(create_urls)} files.")

    if update_files:
        logging.debug(f"Attempting to update {len(update_files)} files.")
        update_files = replication.retry(client.files.update, update_files)
        logging.debug(f"Successfully updated {len(update_files)} files.")

    logging.info(f"Created {len(create_urls)} new files and updated {len(update_files)} existing files.")


def replicate(
    client_src: CogniteClient,
    client_dst: CogniteClient,
    batch_size: int = 10000,
    num_threads: int = 1,
    delete_replicated_if_not_in_src: bool = False,
    delete_not_replicated_in_dst: bool = False,
    skip_unlinkable: bool = False,
    skip_nonasset: bool = False,
):
    """
    Replicates all the files from the source project into the destination project.

    Args:
        client_src: The client corresponding to the source project.
        client_dst: The client corresponding to the destination project.
        batch_size: The biggest batch size to post chunks in.
        num_threads: The number of threads to be used.
        delete_replicated_if_not_in_src: If True, will delete replicated files that are in the destination,
        but no longer in the source project (Default=False).
        delete_not_replicated_in_dst: If True, will delete files from the destination if they were not replicated
        from the source (Default=False).
        skip_unlinkable: If no assets exist in the destination for a file, do not replicate it
        skip_nonasset: If a file has no associated assets, do not replicate it
    """
    project_src = client_src.config.project
    project_dst = client_dst.config.project

    files_src = client_src.files.list(limit=None)
    files_dst = client_dst.files.list(limit=None)
    logging.info(f"There are {len(files_src)} existing files in source ({project_src}).")
    logging.info(f"There are {len(files_dst)} existing files in destination ({project_dst}).")

    src_id_dst_file = replication.make_id_object_map(files_dst)

    assets_dst = client_dst.assets.list(limit=None)
    src_dst_ids_assets = replication.existing_mapping(*assets_dst)
    logging.info(
        f"If a files asset ids is one of the {len(src_dst_ids_assets)} assets "
        f"that have been replicated then it will be linked."
    )

    if skip_unlinkable or skip_nonasset:
        pre_filter_length = len(files_src)
        files_src = replication.filter_objects(files_src, src_dst_ids_assets, skip_unlinkable, skip_nonasset)
        logging.info(f"Filtered out {pre_filter_length - len(files_src)} files. {len(files_src)} files remain.")

    replicated_runtime = int(time.time()) * 1000
    logging.info(f"These copied/updated files will have a replicated run time of: {replicated_runtime}.")

    logging.info(
        f"Starting to copy and update {len(files_src)} files from "
        f"source ({project_src}) to destination ({project_dst})."
    )

    if len(files_src) > batch_size:
        replication.thread(
            num_threads=num_threads,
            copy=copy_files,
            src_objects=files_src,
            src_id_dst_obj=src_id_dst_file,
            src_dst_ids_assets=src_dst_ids_assets,
            project_src=project_src,
            replicated_runtime=replicated_runtime,
            client=client_dst,
        )
    else:
        copy_files(
            src_files=files_src,
            src_id_dst_file=src_id_dst_file,
            src_dst_ids_assets=src_dst_ids_assets,
            project_src=project_src,
            runtime=replicated_runtime,
            client=client_dst,
        )

    logging.info(
        f"Finished copying and updating {len(files_src)} files from "
        f"source ({project_src}) to destination ({project_dst})."
    )

    if delete_replicated_if_not_in_src:
        ids_to_delete = replication.find_objects_to_delete_if_not_in_src(files_src, files_dst)
        client_dst.files.delete(id=ids_to_delete)
        logging.info(
            f"Deleted {len(ids_to_delete)} files in destination ({project_dst})"
            f" because they were no longer in source ({project_src})   "
        )
    if delete_not_replicated_in_dst:
        ids_to_delete = replication.find_objects_to_delete_not_replicated_in_dst(files_dst)
        client_dst.files.delete(id=ids_to_delete)
        logging.info(
            f"Deleted {len(ids_to_delete)} files in destination ({project_dst}) because"
            f"they were not replicated from source ({project_src})   "
        )



import logging
import time
from typing import Dict, List

from cognite.client import CogniteClient
from cognite.client.data_classes import Event

from . import replication


def create_event(src_event: Event, src_dst_ids_assets: Dict[int, int], project_src: str, runtime: int) -> Event:
    """
    Makes a new copy of the event to be replicated based on a source event.

    Args:
        src_event: The event from the source to be replicated to destination.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.

    Returns:
        The replicated event to be created in the destination.
    """
    logging.debug(f"Creating a new event based on source event id {src_event.id}")

    return Event(
        external_id=src_event.external_id,
        start_time=src_event.start_time
        if src_event.start_time and src_event.end_time and src_event.start_time < src_event.end_time
        else src_event.end_time,
        end_time=src_event.end_time,
        type=src_event.type,
        subtype=src_event.subtype,
        description=src_event.description,
        metadata=replication.new_metadata(src_event, project_src, runtime),
        asset_ids=replication.get_asset_ids(src_event.asset_ids, src_dst_ids_assets),
        source=src_event.source,
    )


def update_event(
    src_event: Event, dst_event: Event, src_dst_ids_assets: Dict[int, int], project_src: str, runtime: int
) -> Event:
    """
    Makes an updated version of the destination event based on the corresponding source event.

    Args:
        src_event: The event from the source to be replicated.
        dst_event: The event from the destination that needs to be updated to reflect changes made to its source event.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.

    Returns:
        The updated event object for the replication destination.
    """
    logging.debug(f"Updating existing event {dst_event.id} based on source event id {src_event.id}")

    dst_event.external_id = src_event.external_id
    dst_event.start_time = (
        src_event.start_time
        if src_event.start_time and src_event.end_time and src_event.start_time < src_event.end_time
        else src_event.end_time
    )
    dst_event.end_time = src_event.end_time
    dst_event.type = src_event.type
    dst_event.subtype = src_event.subtype
    dst_event.description = src_event.description
    dst_event.metadata = replication.new_metadata(src_event, project_src, runtime)
    dst_event.asset_ids = replication.get_asset_ids(src_event.asset_ids, src_dst_ids_assets)
    dst_event.source = src_event.source
    return dst_event


def copy_events(
    src_events: List[Event],
    src_id_dst_event: Dict[int, Event],
    src_dst_ids_assets: Dict[int, int],
    project_src: str,
    runtime: int,
    client: CogniteClient,
):
    """
    Creates/updates event objects and then attempts to create and update these objects in the destination.

    Args:
        src_events: A list of the events that are in the source.
        src_id_dst_event:  A dictionary of an events source id to it's matching destination object.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        runtime: The timestamp to be used in the new replicated metadata.
        client: The client corresponding to the destination project.

    """
    logging.debug(f"Starting to replicate {len(src_events)} events.")

    create_events, update_events, unchanged_events = replication.make_objects_batch(
        src_events, src_id_dst_event, src_dst_ids_assets, create_event, update_event, project_src, runtime
    )

    logging.info(f"Creating {len(create_events)} new events and updating {len(update_events)} existing events.")

    if create_events:
        logging.debug(f"Attempting to create {len(create_events)} events.")
        create_events = replication.retry(client.events.create, create_events)
        logging.debug(f"Successfully created {len(create_events)} events.")

    if update_events:
        logging.debug(f"Attempting to update {len(update_events)} events.")
        update_events = replication.retry(client.events.update, update_events)
        logging.debug(f"Successfully updated {len(update_events)} events.")

    logging.info(f"Created {len(create_events)} new events and updated {len(update_events)} existing events.")


def replicate(
    client_src: CogniteClient,
    client_dst: CogniteClient,
    batch_size: int = 10000,
    num_threads: int = 1,
    delete_replicated_if_not_in_src: bool = False,
    delete_not_replicated_in_dst: bool = False,
    skip_unlinkable: bool = False,
    skip_nonasset: bool = False,
):
    """
    Replicates all the events from the source project into the destination project.

    Args:
        client_src: The client corresponding to the source project.
        client_dst: The client corresponding to the destination project.
        batch_size: The biggest batch size to post chunks in.
        num_threads: The number of threads to be used.
        delete_replicated_if_not_in_src: If True, will delete replicated events that are in the destination,
        but no longer in the source project (Default=False).
        delete_not_replicated_in_dst: If True, will delete events from the destination if they were not replicated
        from the source (Default=False).
        skip_unlinkable: If no assets exist in the destination for an event, do not replicate it
        skip_nonasset: If an event has no associated assets, do not replicate it
    """
    project_src = client_src.config.project
    project_dst = client_dst.config.project

    events_src = client_src.events.list(limit=None)
    events_dst = client_dst.events.list(limit=None)
    logging.info(f"There are {len(events_src)} existing events in source ({project_src}).")
    logging.info(f"There are {len(events_dst)} existing events in destination ({project_dst}).")

    src_id_dst_event = replication.make_id_object_map(events_dst)

    assets_dst = client_dst.assets.list(limit=None)
    src_dst_ids_assets = replication.existing_mapping(*assets_dst)
    logging.info(
        f"If an events asset ids is one of the {len(src_dst_ids_assets)} assets "
        f"that have been replicated then it will be linked."
    )

    if skip_unlinkable or skip_nonasset:
        pre_filter_length = len(events_src)
        events_src = replication.filter_objects(events_src, src_dst_ids_assets, skip_unlinkable, skip_nonasset)
        logging.info(f"Filtered out {pre_filter_length - len(events_src)} events. {len(events_src)} events remain.")

    replicated_runtime = int(time.time()) * 1000
    logging.info(f"These copied/updated events will have a replicated run time of: {replicated_runtime}.")

    logging.info(
        f"Starting to copy and update {len(events_src)} events from "
        f"source ({project_src}) to destination ({project_dst})."
    )

    if len(events_src) > batch_size:
        replication.thread(
            num_threads=num_threads,
            copy=copy_events,
            src_objects=events_src,
            src_id_dst_obj=src_id_dst_event,
            src_dst_ids_assets=src_dst_ids_assets,
            project_src=project_src,
            replicated_runtime=replicated_runtime,
            client=client_dst,
        )
    else:
        copy_events(
            src_events=events_src,
            src_id_dst_event=src_id_dst_event,
            src_dst_ids_assets=src_dst_ids_assets,
            project_src=project_src,
            runtime=replicated_runtime,
            client=client_dst,
        )

    logging.info(
        f"Finished copying and updating {len(events_src)} events from "
        f"source ({project_src}) to destination ({project_dst})."
    )

    if delete_replicated_if_not_in_src:
        ids_to_delete = replication.find_objects_to_delete_if_not_in_src(events_src, events_dst)
        client_dst.events.delete(id=ids_to_delete)
        logging.info(
            f"Deleted {len(ids_to_delete)} events in destination ({project_dst})"
            f" because they were no longer in source ({project_src})   "
        )
    if delete_not_replicated_in_dst:
        ids_to_delete = replication.find_objects_to_delete_not_replicated_in_dst(events_dst)
        client_dst.events.delete(id=ids_to_delete)
        logging.info(
            f"Deleted {len(ids_to_delete)} events in destination ({project_dst}) because"
            f"they were not replicated from source ({project_src})   "
        )



import logging
import multiprocessing as mp
import re
from datetime import datetime
from math import ceil, floor
from typing import Any, Callable, List, Optional, Tuple, Union

from cognite.client import CogniteClient
from cognite.client.data_classes import Datapoint, Datapoints
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils._time import timestamp_to_ms


def _get_time_range(src_datapoint: Datapoints, dst_datapoint: Datapoints) -> Tuple[int, int]:
    # +1 because datapoint retrieval time ranges are inclusive on start and exclusive on end
    start_time = 0 if not dst_datapoint else dst_datapoint.timestamp[0] + 1
    end_time = 0 if not src_datapoint else src_datapoint.timestamp[0] + 1
    return start_time, end_time


def _get_chunk(lst: List[Any], num_chunks: int, chunk_number: int) -> List[Any]:
    """Returns a slice of the given list such that all slices are as even in size as possible.

    Args:
        lst: The list to slice
        num_chunks: The amount of chunks that the list should be split into
        chunk_number: Which chunk of the lst to return (0-indexed)

    Returns:
        The chunk_number-th chunk of lst such that the concat of all num_chunks chunks is equivalent to the full lst,
        and each chunk has equal size +-1
    """
    chunk_size = len(lst) // num_chunks
    num_excess_elements = len(lst) % num_chunks

    start_index = chunk_number * chunk_size
    start_index += min(chunk_number, num_excess_elements)  # offset by amount of excess elements used in previous chunks

    end_index = start_index + chunk_size
    if chunk_number < num_excess_elements:  # if we need to include an extra element
        end_index += 1

    return lst[start_index:end_index]


def evaluate_lambda_function(lambda_fnc_str: str):
    """Returns callable object by evaluating lambda function string.

    Args:
        lambda_fnc_str: lambda function string for datapoint.value manipulation

    Returns:
        Callable function

    """
    try:
        lambda_fnc = eval(lambda_fnc_str)
        return lambda_fnc
    except Exception as e:
        logging.error(f"An error occurred when using value manipulation " f"lambda function. {lambda_fnc_str}")
        logging.error(e)
        return None


def replicate_datapoints(
    client_src: CogniteClient,
    client_dst: CogniteClient,
    ts_external_id: str,
    limit: Optional[int] = None,
    partition_size: int = 100000,
    mock_run: bool = False,
    job_id: int = 1,
    src_datapoint_transform: Optional[Callable[[Datapoint], Datapoint]] = None,
    timerange_transform: Optional[Callable[[Tuple[int, int]], Tuple[int, int]]] = None,
    start: Union[int, str] = None,
    end: Union[int, str] = None,
    value_manipulation_lambda_fnc: str = None,
) -> Tuple[bool, int]:
    """
    Copies data points from the source tenant into the destination project, for the given time series.

    If data points already exist in the destination for the time series, only the newer data points in the source are
    copied over.

    Args:
        client_src: The client corresponding to the source project.
        client_dst: The client corresponding to the destination project.
        ts_external_id: The external id of the time series to replicate datapoints for
        limit: The maximum number of data points to copy
        partition_size: The maximum number of datapoints to retrieve per request
        mock_run: If true, only retrieves data points from source and does not insert into destination
        job_id: The batch number being processed
        src_datapoint_transform: Function to apply to all source datapoints before inserting into destination
        timerange_transform: Function to set the time range boundaries (start, end) arbitrarily.
        start: Timestamp to start replication onwards from; if not specified starts at most recent datapoint
        end: If specified, limits replication to datapoints earlier than the end time
        value_manipulation_lambda_fnc: A basic lambda function can be provided to manipulate datapoints as a string.

    Returns:
        A tuple of the success status (True if no failures) and the number of datapoints successfully replicated
    """
    try:
        latest_dst_dp = client_dst.datapoints.retrieve_latest(external_id=ts_external_id)
        latest_src_dp = client_src.datapoints.retrieve_latest(external_id=ts_external_id)
    except CogniteAPIError as exc:
        logging.error(f"Job {job_id}: Failed for external id {ts_external_id}. {exc}")
        return False, 0

    if not latest_src_dp:
        return True, 0

    if src_datapoint_transform:
        latest_src_dp = Datapoints(timestamp=[src_datapoint_transform(latest_src_dp[0]).timestamp])

    _start, _end = _get_time_range(latest_src_dp, latest_dst_dp)

    start = _start if start is None else timestamp_to_ms(start)
    end = _end if end is None else timestamp_to_ms(end)

    if timerange_transform:
        start, end = timerange_transform(start, end)

    # Api Restrictions
    start = max(start, 31536000000)  # 1971

    logging.debug(f"Job {job_id}: Ext_id: {ts_external_id} Retrieving datapoints between {start} and {end}")
    datapoints_count = 0
    while start < end:
        num_to_fetch = partition_size if limit is None else min(partition_size, limit - datapoints_count)
        if num_to_fetch == 0:
            break

        try:
            datapoints = client_src.datapoints.retrieve(
                external_id=ts_external_id, start=start, end=end, limit=num_to_fetch
            )
            if not datapoints:
                break

            if src_datapoint_transform:
                transformed_values = []
                transformed_timestamps = []
                for src_datapoint in datapoints:
                    transformed_datapoint = src_datapoint_transform(src_datapoint)
                    transformed_timestamps.append(transformed_datapoint.timestamp)
                    transformed_values.append(transformed_datapoint.value)
                datapoints = Datapoints(timestamp=transformed_timestamps, value=transformed_values)

            if value_manipulation_lambda_fnc:
                transformed_values = []
                transformed_timestamps = []
                lambda_fnc = evaluate_lambda_function(value_manipulation_lambda_fnc)
                if lambda_fnc:
                    for src_datapoint in datapoints:
                        try:
                            transformed_timestamps.append(src_datapoint.timestamp)
                            transformed_values.append(lambda_fnc(src_datapoint.value))
                        except Exception as e:
                            logging.error(
                                f"Could not manipulate the datapoint (value={src_datapoint.value},"
                                + f" timestamp={src_datapoint.timestamp}). Error: {e}"
                            )
                    datapoints = Datapoints(timestamp=transformed_timestamps, value=transformed_values)

            if not mock_run:
                client_dst.datapoints.insert(datapoints, external_id=ts_external_id)
        except CogniteAPIError as exc:
            logging.error(f"Job {job_id}: Failed for external id {ts_external_id}. {exc}")
            return False, datapoints_count
        else:
            datapoints_count += len(datapoints)
            start = datapoints[-1].timestamp + 1

    logging.debug(f"Job {job_id}: Ext_id: {ts_external_id} Number of datapoints: {datapoints_count}")
    return True, datapoints_count


def batch_replicate(
    client_src: CogniteClient,
    client_dst: CogniteClient,
    job_id: int,
    ext_ids: List[str],
    limit: int,
    mock_run: bool = False,
    partition_size: int = 100000,
    src_datapoint_transform: Optional[Callable[[Datapoint], Datapoint]] = None,
    timerange_transform: Optional[Callable[[Tuple[int, int]], Tuple[int, int]]] = None,
    start: Union[int, str] = None,
    end: Union[int, str] = None,
    value_manipulation_lambda_fnc: str = None,
):
    """
    Replicates datapoints for each time series specified by the external id list.

    Args:
        client_src: The client corresponding to the source project.
        client_dst: The client corresponding to the destination project.
        job_id: The batch number being processed
        ext_ids: The list of external ids for time series to copy over
        limit: The maximum number of data points to copy per time series
        mock_run: If true, only retrieves data points from source and does not insert into destination
        partition_size: The maximum number of datapoints to retrieve per request
        src_datapoint_transform: Function to apply to all source datapoints before inserting into destination
        timerange_transform: Function to set the time range boundaries (start, end) arbitrarily.
        start: Timestamp to start replication onwards from; if not specified starts at most recent datapoint
        end: If specified, limits replication to datapoints earlier than the end time
        value_manipulation_lambda_fnc: A basic lambda function can be provided to manipulate datapoints as a string.
    """

    def log_status(total_ts_count):
        logging.info(
            f"Job {job_id}: Current results: {updated_timeseries_count} time series updated, "
            f"{total_ts_count - updated_timeseries_count - len(failed_external_ids)} "
            f"time series up-to-date. {total_datapoints_copied} datapoints copied. "
            f"{len(failed_external_ids)} failure(s)."
        )

    logging.info(f"Job {job_id}: Starting datapoint replication for {len(ext_ids)} time series...")
    updated_timeseries_count = 0
    total_datapoints_copied = 0
    failed_external_ids = []
    start_time = datetime.now()

    for i, ext_id in enumerate(ext_ids):
        if i % ceil(len(ext_ids) / 10) == 0:
            logging.info(
                f"Job {job_id}: Progress: On time series {i+1}/{len(ext_ids)} "
                f"({floor(100 * i / len(ext_ids))}% complete) in {datetime.now()-start_time}"
            )
            log_status(i)

        success_status, datapoints_copied_count = replicate_datapoints(
            client_src,
            client_dst,
            ext_id,
            partition_size=partition_size,
            mock_run=mock_run,
            job_id=job_id,
            limit=limit,
            src_datapoint_transform=src_datapoint_transform,
            timerange_transform=timerange_transform,
            start=start,
            end=end,
            value_manipulation_lambda_fnc=value_manipulation_lambda_fnc,
        )

        if not success_status:
            failed_external_ids.append(ext_id)
        else:
            updated_timeseries_count += 1
        total_datapoints_copied += datapoints_copied_count

    log_status(len(ext_ids))

    logging.info(f"Total elapsed time: {datetime.now() - start_time}")
    logging.info(f"Job {job_id}: Sample of failed ids: {failed_external_ids[:10]}")


def replicate(
    client_src: CogniteClient,
    client_dst: CogniteClient,
    batch_size: Optional[int] = None,
    num_threads: int = 10,
    limit: Optional[int] = None,
    external_ids: Optional[List[str]] = None,
    mock_run: bool = False,
    partition_size: int = 100000,
    src_datapoint_transform: Optional[Callable[[Datapoint], Datapoint]] = None,
    timerange_transform: Optional[Callable[[Tuple[int, int]], Tuple[int, int]]] = None,
    start: Union[int, str] = None,
    end: Union[int, str] = None,
    exclude_pattern: str = None,
    value_manipulation_lambda_fnc: str = None,
):
    """
    Replicates data points from the source project into the destination project for all time series that
    exist in both environments.

    Args:
        client_src: The client corresponding to the source project.
        client_dst: The client corresponding to the destination project.
        batch_size: The size of batches to split the external id list into. Defaults to num_threads.
        num_threads: The number of threads to be used.
        limit: The maximum number of data points to copy per time series
        external_ids: A list of time series to replicate data points for
        mock_run: If true, runs the replication without insert, printing what would happen
        partition_size: The maximum number of datapoints to retrieve per request
        src_datapoint_transform: Function to apply to all source datapoints before inserting into destination
        timerange_transform: Function to set the time range boundaries (start, end) arbitrarily.
        start: Timestamp to start replication onwards from; if not specified starts at most recent datapoint
        end: If specified, limits replication to datapoints earlier than the end time
        exclude_pattern: Regex pattern; time series whose names match will not be replicated from
        value_manipulation_lambda_fnc: A basic lambda function can be provided to manipulate datapoints as a string.
                                        It will be applied to the value of each datapoint in the timeseries.
    """

    if external_ids and exclude_pattern:
        raise ValueError(
            f"List of time series AND a regex exclusion rule was given! Either remove the filter {exclude_pattern} or the list of time series {external_ids}"
        )
    elif external_ids is not None:  # Specified list of time series is given
        ts_src = client_src.time_series.retrieve_multiple(external_ids=external_ids, ignore_unknown_ids=True)
        ts_dst = client_dst.time_series.retrieve_multiple(external_ids=external_ids, ignore_unknown_ids=True)
        src_ext_id_list = [ts_obj.external_id for ts_obj in ts_src]
    else:
        ts_src = client_src.time_series.list(limit=None)
        ts_dst = client_dst.time_series.list(limit=None)
        filtered_ts_src = []
        skipped_ts = []
        if exclude_pattern:  # Filtering based on regex rule given
            compiled_re = re.compile(exclude_pattern)
            for ts in ts_src:
                if compiled_re.search(ts.external_id):
                    skipped_ts.append(ts.external_id)
                else:
                    filtered_ts_src.append(ts.external_id)
            src_ext_id_list = filtered_ts_src
            logging.info(
                f"Excluding datapoints from {len(skipped_ts)} time series, due to regex rule: {exclude_pattern}. Sample: {skipped_ts[:5]}"
            )
            # Should probably change to logging.debug after a while
        else:  # Expects to replicate all shared time series
            src_ext_id_list = [ts_obj.external_id for ts_obj in ts_src]
    logging.info(f"Number of time series in source: {len(ts_src)}")
    logging.info(f"Number of time series in destination: {len(ts_dst)}")

    dst_ext_id_list = set([ts_obj.external_id for ts_obj in ts_dst])
    shared_external_ids = [ext_id for ext_id in src_ext_id_list if ext_id in dst_ext_id_list and ext_id]
    logging.info(
        f"Number of common time series external ids between destination and source: {len(shared_external_ids)}"
    )

    if batch_size is None:
        batch_size = ceil(len(shared_external_ids) / num_threads)
    num_batches = ceil(len(shared_external_ids) / batch_size)

    arg_list = [
        (
            client_src,
            client_dst,
            job_id,
            _get_chunk(shared_external_ids, num_batches, job_id),
            limit,
            mock_run,
            partition_size,
            src_datapoint_transform,
            timerange_transform,
            start,
            end,
            value_manipulation_lambda_fnc,
        )
        for job_id in range(num_batches)
    ]

    if num_threads > 1:
        with mp.Pool(num_threads) as pool:
            pool.starmap(batch_replicate, arg_list)
    else:
        batch_replicate(*arg_list[0])



import logging
from typing import Dict, List, Tuple, Union

from cognite.client import CogniteClient
from cognite.client.data_classes.raw import Database, Row, Table

from . import replication


def get_not_created_names(
    src_objects: List[Union[Database, Table]], dst_objects: List[Union[Database, Table]]
) -> Tuple[List[str], List[str]]:
    """
    Creates a list of all the source object names and a list of source object names that do not exist in destination.
    Args:
        src_objects: A list of all the source objects, either databases or tables.
        dst_objects: A list of all the destination objects, either databases or tables.

    Returns:
        src_names: A list of all the source object (database or table) names.
        not_created: A list of all the source object (database or table) names that do not exist in destination.

    """

    src_names = [obj.name for obj in src_objects]
    dst_names = [obj.name for obj in dst_objects]

    not_created = [obj_name for obj_name in src_names if obj_name not in dst_names]

    return src_names, not_created


def create_databases_tables(client_src: CogniteClient, client_dst: CogniteClient) -> Dict[str, List[str]]:
    """
    Creates the databases and tables in the destination and makes a dictionary with the database name as the key and
    a list of the table names as the value.

    Args:
        client_src: The client corresponding to the source project.
        client_dst: The client corresponding to the destination project.

    Returns:
        A dictionary mapping database names to a list of it's table names.

    """
    db_list_src = client_src.raw.databases.list(limit=None)
    db_list_dst = client_dst.raw.databases.list(limit=None)

    db_names_src, db_not_created = get_not_created_names(db_list_src, db_list_dst)

    db_create = client_dst.raw.databases.create(name=db_not_created)

    assert len(db_create) == len(db_not_created)

    db_tb = {}
    for db_name in db_names_src:
        tb_list_src = client_src.raw.tables.list(db_name=db_name, limit=None)
        tb_list_dst = client_dst.raw.tables.list(db_name=db_name, limit=None)

        tb_names_src, tb_not_created = get_not_created_names(tb_list_src, tb_list_dst)

        tb_create = client_dst.raw.tables.create(db_name=db_name, name=tb_not_created)

        assert len(tb_create) == len(tb_not_created)

        db_tb[db_name] = tb_names_src

    return db_tb


def insert_rows(rows: List[Row], db_name: str, tb_name: str, client: CogniteClient):
    """
    Inserts the input rows from database and table from the source to the destination.

    Args:
        rows:  A list of all the rows that need to be copied over.
        db_name: The name of the database that rows are copied from.
        tb_name: The name of the table that rows are copied from.
        client: The client corresponding to the destination project.

    Returns:
        The output of insert.

    """
    return client.raw.rows.insert(db_name=db_name, table_name=tb_name, row=rows)


def copy_rows(client_src: CogniteClient, client_dst: CogniteClient, db_tb_dict: Dict[str, List[str]], chunk_size: int):
    """
    Goes through the databases and their tables in order to fetch the rows from source and post them to
    destination in chunks.

    Args:
      client_src: The client corresponding to the source project.
      client_dst: The client corresponding to the destination project.
      db_tb_dict: A dictionary over the databases and tables within the databases.
      chunk_size: The size of the chunks to fetch and post rows in.

    """
    for db_name, tb_list in db_tb_dict.items():
        for tb_name in tb_list:
            for row_list in client_src.raw.rows(db_name=db_name, table_name=tb_name, chunk_size=chunk_size):
                new_rows = {row.key: row.columns for row in row_list}
                kwargs = {"db_name": db_name, "tb_name": tb_name, "client": client_dst}
                replication.retry(insert_rows, new_rows, **kwargs)
            logging.info(f"Rows are added from database {db_name} and table {tb_name}")


def replicate(client_src: CogniteClient, client_dst: CogniteClient, chunk_size: int):
    """
    Replicates all the raw from the source project into the destination project.

    Args:
        client_src: The client corresponding to the source project.
        client_dst: The client corresponding to the destination project.
        chunk_size: The biggest chunk size to fetch and post in.

    """
    project_src = client_src.config.project
    project_dst = client_dst.config.project

    logging.info(f"Starting to copy raw from source ({project_src}) to destination ({project_dst}).")

    db_tb = create_databases_tables(client_src, client_dst)
    logging.info(
        f"There are {len(db_tb)} databases with a total of {sum([len(db_tb[x]) for x in db_tb])} "
        f"tables to be copied to the destination"
    )

    copy_rows(client_src, client_dst, db_tb, chunk_size)

    logging.info(f"Finished copying raw from source ({project_src}) to destination ({project_dst}).")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cognite', 'cognite.replicator']

package_data = \
{'': ['*']}

install_requires = \
['cognite-sdk-core>=1.4.10,<2.0.0',
 'google-cloud-logging>=1.12,<2.0',
 'pyyaml>=5.1.0,<6.0.0']

entry_points = \
{'console_scripts': ['replicator = cognite.replicator.__main__:main']}

setup_kwargs = {
    'name': 'cognite-replicator',
    'version': '0.8.1',
    'description': 'Python package for replicating data across CDF tenants. Copyright 2019 Cognite AS',
    'long_description': '<a href="https://cognite.com/">\n    <img src="https://raw.githubusercontent.com/cognitedata/cognite-python-docs/master/img/cognite_logo.png" alt="Cognite logo" title="Cognite" align="right" height="80" />\n</a>\n\n# Cognite Python Replicator\n[![build](https://webhooks.dev.cognite.ai/build/buildStatus/icon?job=github-builds/cognite-replicator/master)](https://jenkins.cognite.ai/job/github-builds/job/cognite-replicator/job/master/)\n[![codecov](https://codecov.io/gh/cognitedata/cognite-replicator/branch/master/graph/badge.svg)](https://codecov.io/gh/cognitedata/cognite-replicator)\n[![Documentation Status](https://readthedocs.com/projects/cognite-cognite-replicator/badge/?version=latest)](https://cognite-cognite-replicator.readthedocs-hosted.com/en/latest/)\n[![PyPI version](https://badge.fury.io/py/cognite-replicator.svg)](https://pypi.org/project/cognite-replicator/)\n[![tox](https://img.shields.io/badge/tox-3.6%2B-blue.svg)](https://www.python.org/downloads/release/python-366/)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/cognite-replicator)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\nCognite Replicator is a Python package for replicating data across Cognite Data Fusion (CDF) projects. This package is\nbuilt on top of the Cognite Python SDK.\n\nCopyright 2019 Cognite AS\n\n## Prerequisites\nIn order to start using the Replicator, you need:\n* Python3 (>= 3.6)\n* Two API keys, one for your source tenant and one for your destination tenant. Never include the API key directly in the code or upload the key to github. Instead, set the API key as an environment variable.\n\nThis is how you set the API key as an environment variable on Mac OS and Linux:\n```bash\n$ export COGNITE_SOURCE_API_KEY=<your source API key>\n$ export COGNITE_DESTINATION_API_KEY=<your destination API key>\n```\n\n## Installation\nThe replicator is available on [PyPI](https://pypi.org/project/cognite-replicator/), and can also be executed as a standalone script.\n\nTo run it from command line, run:\n```bash\npip install cognite-replicator\npython -m cognite.replicator config/filepath.yml\n```\nIf no file is specified then replicator will use config/default.yml.\n\nAlternatively, build and run it as a docker container. The image is avaible on [docker hub](https://hub.docker.com/r/cognite/cognite-replicator):\n```bash\ndocker build -t cognite-replicator .\ndocker run -it cognite-replicator\n```\n\nFor *Databricks* you can install it on a cluster. First, click on **Libraries** and **Install New**.  Choose your library type to be **PyPI**, and enter **cognite-replicator** as Package. Let the new library install and you are ready to replicate!\n\n\n## Usage\n\n### Setup as Python library\n```python\nimport os\n\nfrom cognite.client import CogniteClient\nfrom cognite.replicator import assets, events, time_series, datapoints\n\nSRC_API_KEY = os.environ.get("COGNITE_SOURCE_API_KEY")\nDST_API_KEY = os.environ.get("COGNITE_DESTINATION_API_KEY")\nPROJECT_SRC = "Name of source tenant"\nPROJECT_DST = "Name of destination tenant"\nCLIENT_NAME = "cognite-replicator"\nBATCH_SIZE = 10000 # this is the max size of a batch to be posted\nNUM_THREADS= 10 # this is the max number of threads to be used\n\nCLIENT_SRC = CogniteClient(api_key=SRC_API_KEY, project=PROJECT_SRC, client_name=CLIENT_NAME)\nCLIENT_DST = CogniteClient(api_key=DST_API_KEY, project=PROJECT_DST, client_name=CLIENT_NAME, timeout=90)\n\nassets.replicate(CLIENT_SRC, CLIENT_DST)\nevents.replicate(CLIENT_SRC, CLIENT_DST, BATCH_SIZE, NUM_THREADS)\ntime_series.replicate(CLIENT_SRC, CLIENT_DST, BATCH_SIZE, NUM_THREADS)\ndatapoints.replicate(CLIENT_SRC, CLIENT_DST)\n```\n\n### Run it from databricks notebook\n```python\nimport logging\n\nfrom cognite.client import CogniteClient\nfrom cognite.replicator import assets, configure_databricks_logger\n\nSRC_API_KEY = dbutils.secrets.get("cdf-api-keys", "source-tenant")\nDST_API_KEY = dbutils.secrets.get("cdf-api-keys", "destination-tenant")\n\nCLIENT_SRC = CogniteClient(api_key=SRC_API_KEY, client_name="cognite-replicator")\nCLIENT_DST = CogniteClient(api_key=DST_API_KEY, client_name="cognite-replicator")\n\nconfigure_databricks_logger(log_level=logging.INFO)\nassets.replicate(CLIENT_SRC, CLIENT_DST)\n```\n\n## Changelog\nWondering about upcoming or previous changes? Take a look at the [CHANGELOG](https://github.com/cognitedata/cognite-replicator/blob/master/CHANGELOG.md).\n\n## Contributing\nWant to contribute? Check out [CONTRIBUTING](https://github.com/cognitedata/cognite-replicator/blob/master/CONTRIBUTING.md).\n',
    'author': 'Even Wiik Thomassen',
    'author_email': 'even.thomassen@cognite.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
"""
A tool for replicating data from one CDF tenant to another.

To run, configure the source and destination CogniteClient with project and api_key.
API keys can be set as environment variables COGNITE_SOURCE_API_KEY and
COGNITE_DESTINATION_API_KEY or through command line arguments.

You must provide a config file, give the path either in environment variable
COGNITE_CONFIG_FILE or as command line argument.

Example usage: poetry run replicator
"""
import argparse
import logging
import os
import sys
from enum import Enum, auto, unique
from pathlib import Path
from typing import Optional

import yaml

from cognite.client import CogniteClient
from cognite.client.exceptions import CogniteAPIError

from . import assets, configure_logger, datapoints, events, files, raw, time_series

ENV_VAR_FOR_CONFIG_FILE_PATH = "COGNITE_CONFIG_FILE"


@unique
class Resource(Enum):
    """CDF Resource types that can be replicated."""

    ALL = auto()
    ASSETS = auto()
    EVENTS = auto()
    RAW = auto()
    TIMESERIES = auto()
    DATAPOINTS = auto()
    FILES = auto()


def create_cli_parser() -> argparse.ArgumentParser:
    """Returns ArgumentParser for command line interface."""
    parser = argparse.ArgumentParser()
    parser.add_argument("config", nargs="?", help="path to yaml configuration file")
    return parser


def _validate_login(src_client: CogniteClient, dst_client: CogniteClient, src_project: str, dst_project: str) -> bool:
    """Login with CogniteClients and validate projects if set."""
    try:
        src_login_status = src_client.login.status()
        dst_login_status = dst_client.login.status()
    except CogniteAPIError as exc:
        logging.fatal("Failed to login with CogniteClient {!s}".format(exc))
        return False
    if src_project and src_login_status.project != src_project:
        logging.fatal("Source project don't match with API key configuration")
        return False
    if dst_project and dst_login_status.project != dst_project:
        logging.fatal("Destination project don't match with API key configuration")
        return False
    return True


def _get_config_path(config_arg: Optional[str]) -> Path:
    """Get the config file, first either from given path or from env variable."""
    if config_arg:
        config_file = Path(config_arg)
    elif os.environ.get(ENV_VAR_FOR_CONFIG_FILE_PATH):
        config_file = Path(os.environ[ENV_VAR_FOR_CONFIG_FILE_PATH])
    else:
        config_file = None

    if not config_file or not config_file.is_file():
        logging.fatal(f"Config file not found: {config_file}")
        sys.exit(1)
    return config_file


def main():
    args = create_cli_parser().parse_args()
    with open(_get_config_path(args.config)) as config_file:
        config = yaml.safe_load(config_file.read())

    configure_logger(config.get("log_level", "INFO").upper(), Path(config.get("log_path", "log")))

    delete_replicated_if_not_in_src = config.get("delete_if_removed_in_source", False)
    delete_not_replicated_in_dst = config.get("delete_if_not_replicated", False)

    src_api_key = os.environ.get(config.get("src_api_key_env_var", "COGNITE_SOURCE_API_KEY"))
    dst_api_key = os.environ.get(config.get("dst_api_key_env_var", "COGNITE_DESTINATION_API_KEY"))

    src_client = CogniteClient(
        api_key=src_api_key,
        project=config.get("src_project"),
        client_name=config.get("client_name"),
        base_url=config.get("src_baseurl", "https://api.cognitedata.com"),
        timeout=config.get("client_timeout"),
    )
    dst_client = CogniteClient(
        api_key=dst_api_key,
        project=config.get("dst_project"),
        client_name=config.get("client_name"),
        base_url=config.get("dst_baseurl", "https://api.cognitedata.com"),
        timeout=config.get("client_timeout"),
    )

    if not _validate_login(src_client, dst_client, config.get("src_project"), config.get("dst_project")):
        sys.exit(2)

    resources_to_replicate = {Resource[resource.upper()] for resource in config.get("resources")}
    if Resource.ALL in resources_to_replicate:
        resources_to_replicate.update({resource for resource in Resource})

    if Resource.ASSETS in resources_to_replicate:
        assets.replicate(
            src_client,
            dst_client,
            delete_replicated_if_not_in_src=delete_replicated_if_not_in_src,
            delete_not_replicated_in_dst=delete_not_replicated_in_dst,
        )

    if Resource.EVENTS in resources_to_replicate:
        events.replicate(
            src_client,
            dst_client,
            config.get("batch_size"),
            config.get("number_of_threads"),
            delete_replicated_if_not_in_src=delete_replicated_if_not_in_src,
            delete_not_replicated_in_dst=delete_not_replicated_in_dst,
        )

    if Resource.TIMESERIES in resources_to_replicate:
        time_series.replicate(
            src_client,
            dst_client,
            config.get("batch_size"),
            config.get("number_of_threads"),
            delete_replicated_if_not_in_src=delete_replicated_if_not_in_src,
            delete_not_replicated_in_dst=delete_not_replicated_in_dst,
            target_external_ids=config.get("timeseries_external_ids"),
            exclude_pattern=config.get("timeseries_exclude_pattern"),
        )

    if Resource.FILES in resources_to_replicate:
        files.replicate(
            src_client,
            dst_client,
            config.get("batch_size"),
            config.get("number_of_threads"),
            delete_replicated_if_not_in_src=delete_replicated_if_not_in_src,
            delete_not_replicated_in_dst=delete_not_replicated_in_dst,
        )

    if Resource.RAW in resources_to_replicate:
        raw.replicate(src_client, dst_client, config.get("batch_size"))

    if Resource.DATAPOINTS in resources_to_replicate:
        datapoints.replicate(
            client_src=src_client,
            client_dst=dst_client,
            batch_size=config.get("batch_size_datapoints"),
            num_threads=config.get("number_of_threads"),
            limit=config.get("datapoint_limit"),
            external_ids=config.get("timeseries_external_ids"),
            start=config.get("datapoints_start"),
            end=config.get("datapoints_end"),
            exclude_pattern=config.get("timeseries_exclude_pattern"),
            value_manipulation_lambda_fnc=config.get("value_manipulation_lambda_fnc"),
        )


if __name__ == "__main__":
    main()



import logging
import threading
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import requests

from cognite.client import CogniteClient
from cognite.client.data_classes import Event, FileMetadata, TimeSeries
from cognite.client.data_classes.assets import Asset
from cognite.client.data_classes.raw import Row


def make_id_object_map(
    objects: List[Union[Asset, Event, FileMetadata, TimeSeries]]
) -> Dict[int, Union[Asset, Event, FileMetadata, TimeSeries]]:
    """
    Makes a dictionary with the source object id as the key and the object as the value for objects
    that have been replicated.

    Args:
        objects: A list of objects that are from the replication destination.

    Returns:
        A dictionary of source object id to destination object for objects that have been replicated.
    """
    return {
        int(obj.metadata["_replicatedInternalId"]): obj
        for obj in objects
        if obj.metadata and obj.metadata.get("_replicatedInternalId")
    }


def filter_objects(
    objects: Union[List[Event], List[FileMetadata], List[TimeSeries]],
    src_dst_ids_assets: Dict[int, int],
    skip_unlinkable: bool = False,
    skip_nonasset: bool = False,
    filter_fn: Optional[Callable[[Union[Event, FileMetadata, TimeSeries]], bool]] = None,
) -> Union[List[Event], List[FileMetadata], List[TimeSeries]]:
    """Filters out objects based on their assets and optionally custom filter logic

    Args:
        objects: A list of all the objects to filter.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        skip_unlinkable: If True, excludes objects whose assets haven't been replicated in the destination
        skip_nonasset: If True, excludes objects without any associated assets
        filter_fn: If specified, is used to filter the objects in addition to the existing asset filters.

    Returns:
        A list of objects that meet the criteria.
    """
    filtered_objs = []

    def has_assets(obj):
        is_multi_asset = hasattr(obj, "asset_ids")
        return (is_multi_asset and getattr(obj, "asset_ids", None) is not None) or (
            not is_multi_asset and getattr(obj, "asset_id", None) is not None
        )

    for object in objects:
        if filter_fn is not None and not filter_fn(object):
            continue

        if skip_nonasset and not has_assets(object):
            continue

        if skip_unlinkable and has_assets(object):
            linkable = False
            asset_id_list = getattr(object, "asset_ids", [getattr(object, "asset_id", -1)])
            for asset_id in asset_id_list:
                if asset_id in src_dst_ids_assets:
                    linkable = True
                    break

            if not linkable:
                continue
        filtered_objs.append(object)

    return filtered_objs


def existing_mapping(*objects: List[Asset], ids: Dict[int, int] = None) -> Dict[int, int]:
    """
    Updates a dictionary with all the source id to destination id pairs for the objects that have been replicated.

    Args:
        *objects: A list of objects to make a mapping of.
        ids: A dictionary of all the mappings of source object id to destination object id.

    Returns:
        The updated dictionary with the ids from new objects that have been replicated.
    """
    if not ids:
        ids = {}

    for obj in objects:
        if obj.metadata and obj.metadata.get("_replicatedInternalId"):
            ids[int(obj.metadata["_replicatedInternalId"])] = obj.id

    return ids


def get_asset_ids(ids: List[int], src_dst_ids_assets: Dict[int, int]) -> Union[List[int], None]:
    """
    Create the list of destination asset ids from the list of source asset ids.

    Args:
        ids: A list of the source asset ids from the source object.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.

    Returns:
        A list of the destination asset ids for the destination object.
    """
    if ids is None:
        return None

    return [src_dst_ids_assets[src_asset_id] for src_asset_id in ids if src_asset_id in src_dst_ids_assets]


def new_metadata(
    obj: Union[Asset, Event, FileMetadata, TimeSeries], project_src: str, replicated_runtime: int
) -> Dict[str, Union[int, str]]:
    """
    Copies the objects metadata and adds three new fields to it providing information about the objects replication.

    Fields Created
        - **_replicatedSource**: The name of the project this object is replicated from.
        - **_replicatedTime**: The timestamp of when the object was replicated, all objects created/updated in the same
          execution will have the same timestamp.
        - **_replicatedInternalId**: The internal id of the source object that the destination object
          is being replicated from.

    Args:
        obj: The source object that is being replicated to the destination.
        project_src: The name of the project the object is being replicated from.
        replicated_runtime: The timestamp to be used in the new replicated metadata.

    Returns:
        The metadata dictionary for the replicated destination object based on the source object.

    """
    metadata: Dict[str, Any] = dict(obj.metadata if obj.metadata else {})
    metadata["_replicatedSource"] = project_src
    metadata["_replicatedTime"] = replicated_runtime
    metadata["_replicatedInternalId"] = obj.id
    return metadata


def make_objects_batch(
    src_objects: List[Union[Asset, Event, FileMetadata, TimeSeries]],
    src_id_dst_map: Dict[int, Union[Asset, Event, FileMetadata, TimeSeries]],
    src_dst_ids_assets: Dict[int, int],
    create,
    update,
    project_src: str,
    replicated_runtime: int,
    depth: Optional[int] = None,
    dst_ts: Optional[TimeSeries] = None,
) -> Tuple[
    List[Union[Asset, Event, FileMetadata, TimeSeries]],
    List[Union[Asset, Event, FileMetadata, TimeSeries]],
    List[Union[Asset, Event, FileMetadata, TimeSeries]],
]:
    """
    Create a batch of new objects from a list of source objects or update existing destination objects to their
    corresponding source object.

    Args:
        src_objects: A list of objects to be replicated from a source.
        src_id_dst_map: A dictionary of source object ids to the matching destination object.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        create: The function to be used in order to create all the objects in CDF.
        update: The function to be used in order to update the existing objects in CDF.
        project_src: The name of the project the object is being replicated from.
        replicated_runtime: The timestamp to be used in the new replicated metadata.
        depth: The depth of the asset within the asset hierarchy, only used for making assets.
        dst_ts: List of timeseries in the destination - Will be used for comparison if current timeseries where not copied by the replicator
    Returns:
        create_objects: A list of all the new objects to be posted to CDF.
        update_objects: A list of all the updated objects to be updated in CDF.
        unchanged_objects: A list of all the objects that don't need to be updated.
    """

    create_objects = []
    update_objects = []
    unchanged_objects = []

    kwargs = {"depth": depth} if depth is not None else {}  # Only used on assets

    # make a set of external ids to loop through
    if dst_ts:
        dst_ts_ext_id_set = {ts.external_id for ts in dst_ts}
    else:
        dst_ts_ext_id_set = set()

    for src_obj in src_objects:
        dst_obj = src_id_dst_map.get(src_obj.id)
        if dst_obj:
            if not src_obj.last_updated_time or src_obj.last_updated_time > int(dst_obj.metadata["_replicatedTime"]):
                dst_obj = update(src_obj, dst_obj, src_dst_ids_assets, project_src, replicated_runtime, **kwargs)
                update_objects.append(dst_obj)
            else:
                unchanged_objects.append(dst_obj)
        elif dst_ts:
            if src_obj.external_id in dst_ts_ext_id_set:
                unchanged_objects.append(src_obj)
            else:
                new_asset = create(src_obj, src_dst_ids_assets, project_src, replicated_runtime, **kwargs)
                create_objects.append(new_asset)
        else:
            new_asset = create(src_obj, src_dst_ids_assets, project_src, replicated_runtime, **kwargs)
            create_objects.append(new_asset)

    return create_objects, update_objects, unchanged_objects


def retry(
    function, objects: List[Union[Asset, Event, FileMetadata, TimeSeries, Row]], **kwargs
) -> List[Union[Asset, Event, TimeSeries]]:
    """
    Attempt to either create/update the objects, if it fails retry creating/updating the objects. This will retry up
    to three times.

    Args:
        function: The function that will be applied to objects, either creating_objects or updating_objects.
        objects: A list of all the new objects or updated objects.

    Returns:
        A list of all the objects that were created or updated in CDF.

    """
    ret: List[Union[Asset, Event, FileMetadata, TimeSeries, Row]] = []
    if objects:
        tries = 3
        for i in range(tries):
            logging.info("Current try: %d" % i)
            try:
                ret = function(objects, **kwargs)
                break
            except requests.exceptions.ReadTimeout as e:
                logging.warning(f"Retrying due to {e}")

    return ret


def thread(
    num_threads: int,
    copy,
    src_objects: List[Union[Event, FileMetadata, TimeSeries]],
    src_id_dst_obj: Dict[int, Union[Event, FileMetadata, TimeSeries]],
    src_dst_ids_assets: Dict[int, int],
    project_src: str,
    replicated_runtime: int,
    client: CogniteClient,
    dst_ts: Optional[TimeSeries] = None,
):
    """
    Split up objects to replicate them in batches and thread each batch.

    Args:
        num_threads: The number of threads to be used.
        copy: The function used to copy objects.
        src_objects: A list of all the objects in the source to be replicated.
        src_id_dst_obj: A dictionary of source object id to destination object.
        src_dst_ids_assets: A dictionary of all the mappings of source asset id to destination asset id.
        project_src: The name of the project the object is being replicated from.
        replicated_runtime: The timestamp to be used in the new replicated metadata.
        client: The Cognite Client for the destination project.
        dst_ts: List of timeseries in the destination - Will be used for comparison if current timeseries where not copied by the replicator

    """

    threads = []
    chunk_size = len(src_objects) // num_threads
    logging.info(f"Thread chunk size: {chunk_size}")
    remainder = len(src_objects) % num_threads
    logging.info(f"Thread remainder size: {remainder}\n")

    i = 0
    for t in range(num_threads):
        cs = chunk_size + int(t < remainder)
        threads.append(
            threading.Thread(
                target=copy,
                args=(
                    src_objects[i : i + cs],
                    src_id_dst_obj,
                    src_dst_ids_assets,
                    project_src,
                    replicated_runtime,
                    client,
                    dst_ts,
                ),
            )
        )
        i += cs

    assert i == len(src_objects)

    for count, t in enumerate(threads, start=1):
        t.start()
        logging.info(f"Started thread: {count}")

    for count, t in enumerate(threads, start=1):
        t.join()
        logging.info(f"Joined thread: {count}")


def remove_replication_metadata(objects: Union[List[Asset], List[Event], List[TimeSeries]]):
    """Removes the replication metadata from the passed resource list, so that the resources will look original.
    See also clear_replication_metadata.

    Parameters:
        objects: The list of objects to strip replication metadata away from
    """
    for obj in objects:
        if obj.metadata:
            obj.metadata.pop("_replicatedInternalId", None)
            obj.metadata.pop("_replicatedSource", None)
            obj.metadata.pop("_replicatedTime", None)


def clear_replication_metadata(client: CogniteClient):
    """Removes the replication metadata from all resources, so that the replicated tenant will look like an original.
    Sample use-case: clean up a demo-tenant so that extra data is not present.

    Caution: After clearing replication metadata, the delete_replicated_if_not_in_src option will delete everything.

    Parameters:
        client: The client to strip replication metadata away from
    """

    logging.info("Starting to clear replication metadata...")
    events_dst = client.events.list(limit=None)
    ts_dst = client.time_series.list(limit=None)
    assets_dst = client.assets.list(limit=None)
    client.events.update(events_dst)
    client.time_series.update(ts_dst)
    client.assets.update(assets_dst)
    logging.info("Replication metadata cleared!")


def find_objects_to_delete_not_replicated_in_dst(
    dst_objects: List[Union[Asset, Event, FileMetadata, TimeSeries]]
) -> List[Asset]:
    """
    Deleting all the assets in the destination that do not have the "_replicatedSource" in metadata, which
    means that is was not copied from the source, but created in the destination.

    Parameters:
        dst_objects: The list of objects from the dst destination.
    """
    obj_ids_to_remove = []
    for obj in dst_objects:
        if not obj.metadata or not obj.metadata.get("_replicatedSource"):
            obj_ids_to_remove.append(obj.id)
    return obj_ids_to_remove


def find_objects_to_delete_if_not_in_src(
    src_objects: List[Union[Asset, Event, FileMetadata, TimeSeries]],
    dst_objects: List[Union[Asset, Event, FileMetadata, TimeSeries]],
) -> List[Union[Asset, Event, FileMetadata, TimeSeries]]:
    """
    Compare the destination and source assets and delete the ones that are no longer in the source.

    Parameters:
        src_objects: The list of objects from the src destination.
        dst_objects: The list of objects from the dst destination.
    """

    src_obj_ids = {obj.id for obj in src_objects}

    obj_ids_to_remove = []
    for obj in dst_objects:
        if obj.metadata and obj.metadata.get("_replicatedInternalId"):
            if int(obj.metadata["_replicatedInternalId"]) not in src_obj_ids:
                obj_ids_to_remove.append(obj.id)
    return obj_ids_to_remove
